@@ -4,12 +4,13 @@ import os
 import socket
 from stat import S_IMODE, S_ISDIR, S_ISREG
 import tempfile
+
 import paramiko
 from paramiko import SSHException   # make available
 from paramiko import AuthenticationException   # make available
 from paramiko import AgentKey
 
-__version__ = "0.2.6"
+__version__ = "0.2.7"
 
 
 def st_mode_to_int(val):
@@ -50,6 +51,40 @@ class CredentialException(Exception):
         Exception.__init__(self, message)
         self.message = message
 
+class WTCallbacks(object):
+    '''an object to house the callbacks, used internally
+
+    :ivar flist: list of files currently traversed
+    :ivar dlist: list of directories currently traversed
+    :ivar ulist: list of unknown entities currently traversed
+    '''
+    def __init__(self):
+        '''set instance vars'''
+        self.flist = []
+        self.dlist = []
+        self.ulist = []
+
+    def file_cb(self, pathname):
+        '''called for regular files
+
+        :param str pathname: file path
+        '''
+        self.flist.append(pathname)
+
+    def dir_cb(self, pathname):
+        '''called for directories
+
+        :param str pathname: directory path
+        '''
+        self.dlist.append(pathname)
+
+    def unk_cb(self, pathname):
+        '''called for unknown file types
+
+        :param str pathname: unknown entity path
+        '''
+        self.ulist.append(pathname)
+
 
 class Connection(object):
     """Connects and logs into the specified hostname.
@@ -67,7 +102,7 @@ class Connection(object):
         pysftp creates a temporary file and logs to that.  If set to a valid
         path and filename, pysftp logs to that.  The name of the logfile can
         be found at  ``.logfile``
-    :returns: a connection to the requested host
+    :returns: (obj) connection to the requested host
     :raises:
         ConnectionException, CredentialException, SSHException,
         AuthenticationException, PasswordRequiredException
@@ -89,7 +124,6 @@ class Connection(object):
         if not username:
             username = os.environ['LOGNAME']
 
-
         self._logfile = log
         if log:
             if isinstance(log, bool):
@@ -146,6 +180,16 @@ class Connection(object):
             self._sftp = paramiko.SFTPClient.from_transport(self._transport)
             self._sftp_live = True
 
+    @property
+    def pwd(self):
+        '''return the current working directory
+
+        :returns: (str) current working directory
+
+        '''
+        self._sftp_connect()
+        return self._sftp.normalize('.')
+
     def get(self, remotepath, localpath=None, callback=None,
             preserve_mtime=False):
         """Copies a file between the remote host and the local host.
@@ -153,7 +197,7 @@ class Connection(object):
         :param str remotepath: the remote path and filename, source
         :param str localpath:
             the local path and filename to copy, destination. If not specified,
-            file is copied to local cwd
+            file is copied to local current working directory
         :param callable callback:
             optional callback function (form: ``func(int, int)``) that accepts
             the bytes transferred so far and the total bytes to be transferred.
@@ -162,7 +206,7 @@ class Connection(object):
             local file match the time on the remote. (st_atime can differ
             because stat'ing the localfile can/does update it's st_atime)
 
-        :returns: nothing
+        :returns: None
 
         :raises: IOError
 
@@ -178,6 +222,69 @@ class Connection(object):
         if preserve_mtime:
             os.utime(localpath, (sftpattrs.st_atime, sftpattrs.st_mtime))
 
+    def get_d(self, remotedir, localdir, preserve_mtime=False):
+        """get the contents of remotedir and write to locadir. (non-recursive)
+
+        :param str remotedir: the remote directory to copy from (source)
+        :param str localdir: the local directory to copy to (target)
+        :param bool preserve_mtime:
+            preserve modification time on files(default: False)
+
+        :returns: None
+
+        :raises:
+        """
+        self._sftp_connect()
+        saved = self.pwd
+        self.cwd(remotedir)
+        for sattr in self._sftp.listdir_attr('.'):
+            if S_ISREG(sattr.st_mode):
+                rname = sattr.filename
+                self.get(rname, reparent(localdir, rname),
+                         preserve_mtime=preserve_mtime)
+        self.cwd(saved)
+
+    def get_r(self, remotedir, localdir, preserve_mtime=False):
+        """recursively copy remotedir structure to localdir
+
+        :param str remotedir: the remote directory to copy from
+        :param str localdir: the local directory to copy to
+        :param bool preserve_mtime:
+            preserve modification time on files(default: False)
+
+        :returns: None
+
+        :raises:
+
+        """
+        self._sftp_connect()
+        wtcb = WTCallbacks()
+        self.walktree(remotedir, wtcb.file_cb, wtcb.dir_cb, wtcb.unk_cb)
+        # handle directories we recursed through
+        for dname in wtcb.dlist:
+            for subdir in path_advance(dname):
+                try:
+                    os.mkdir(reparent(localdir, subdir))
+                    wtcb.dlist.append(subdir)
+                except OSError:     # dir exists
+                    pass
+
+        for fname in wtcb.flist:
+            # they may have told us to start down farther, so we may not have
+            # recursed through some, ensure local dir structure matches
+            head, _ = os.path.split(fname)
+            if head not in wtcb.dlist:
+                for subdir in path_advance(head):
+                    if subdir not in wtcb.dlist and subdir != '.':
+                        os.mkdir(reparent(localdir, subdir))
+                        wtcb.dlist.append(subdir)
+
+            self.get(fname,
+                     reparent(localdir, fname),
+                     preserve_mtime=preserve_mtime
+                    )
+
+
     def getfo(self, remotepath, flo, callback=None):
         """Copy a remote file (remotepath) to a file-like object, flo.
 
@@ -187,7 +294,7 @@ class Connection(object):
             optional callback function (form: ``func(int, int``)) that accepts
             the bytes transferred so far and the total bytes to be transferred.
 
-        :returns int: the number of bytes written to the opened file object
+        :returns: (int) the number of bytes written to the opened file object
 
         :raises: Any exception raised by operations will be passed through.
 
@@ -201,7 +308,7 @@ class Connection(object):
 
         :param str localpath: the local path and filename
         :param str remotepath:
-            the remote path, else the remote cwd() and filename is used.
+            the remote path, else the remote :attr:`.pwd` and filename is used.
         :param callable callback:
             optional callback function (form: ``func(int, int``)) that accepts
             the bytes transferred so far and the total bytes to be transferred..
@@ -214,7 +321,7 @@ class Connection(object):
             because stat'ing the localfile can/does update it's st_atime)
 
         :returns:
-            SFTPAttributes object containing attributes about the given file
+            (obj) SFTPAttributes containing attributes about the given file
 
         :raises: IOError, OSError
 
@@ -253,9 +360,9 @@ class Connection(object):
             size
 
         :returns:
-            SFTPAttributes object containing attributes about the given file
+            (obj) SFTPAttributes containing attributes about the given file
 
-        :raises: TypeError if remotepath not specified, any underlying error
+        :raises: TypeError, if remotepath not specified, any underlying error
 
         """
         self._sftp_connect()
@@ -264,11 +371,11 @@ class Connection(object):
 
     def execute(self, command):
         """Execute the given commands on a remote machine.  The command is
-        executed without regard to the remote cwd.
+        executed without regard to the remote :attr:`.pwd`.
 
         :param str command: the command to execute.
 
-        :returns: results
+        :returns: (list of str) representing the results of the command
 
         :raises: Any exception raised by command will be passed through.
 
@@ -286,14 +393,16 @@ class Connection(object):
 
         :param str remotepath: the remote path to change to
 
-        :returns: nothing
+        :returns: None
 
-        :raises: IOError
+        :raises: IOError, if path does not exist
 
         """
         self._sftp_connect()
         self._sftp.chdir(remotepath)
 
+    cwd = chdir     # synonym for chdir
+
     def chmod(self, remotepath, mode=777):
         """set the mode of a remotepath to mode, where mode is an integer
         representation of the octal mode to use.
@@ -304,7 +413,7 @@ class Connection(object):
 
         :returns: None
 
-        :raises: IOError if the file doesn't exist
+        :raises: IOError, if the file doesn't exist
 
         """
         self._sftp_connect()
@@ -319,9 +428,9 @@ class Connection(object):
         :param int uid: the user id to set on the remotepath
         :param int gid: the group id to set on the remotepath
 
-        :returns None:
+        :returns: None
 
-        :raises IOError: if you don't have permission or the file doesn't exist
+        :raises: IOError, if you don't have permission or the file doesn't exist
 
         """
         self._sftp_connect()
@@ -337,24 +446,44 @@ class Connection(object):
         self._sftp.chown(remotepath, uid=uid, gid=gid)
 
     def getcwd(self):
-        """return the current working directory on the remote
+        """return the current working directory on the remote. This is a wrapper
+        for paramiko's method and not to be confused with the SFTP command, cwd.
 
-        :returns: a string representing the current remote path
+        :returns: (str) the current remote path. None, if not set.
 
         """
         self._sftp_connect()
         return self._sftp.getcwd()
 
     def listdir(self, remotepath='.'):
-        """return a list of files/directories for the given remote path
+        """return a list of files/directories for the given remote path.
+        Unlike, paramiko, the directory listing is sorted.
 
         :param str remotepath: path to list on the server
 
-        :returns: a list of entries
+        :returns: (list) directory entries, sorted
 
         """
         self._sftp_connect()
-        return self._sftp.listdir(remotepath)
+        return sorted(self._sftp.listdir(remotepath))
+
+    def listdir_attr(self, remotepath='.'):
+        """return a list of SFTPAttribute objects of the files/directories for
+        the given remote path. The list is in arbitrary order. It does not
+        include the special entries '.' and '..'.
+
+        The returned SFTPAttributes objects will each have an additional field:
+        longname, which may contain a formatted string of the file's
+        attributes, in unix format. The content of this string will depend on
+        the SFTP server.
+
+        :param str remotepath: path to list on the server
+
+        :returns: (list of SFTPAttributes)
+
+        """
+        self._sftp_connect()
+        return self._sftp.listdir_attr(remotepath)
 
     def mkdir(self, remotepath, mode=777):
         """Create a directory named remotepath with mode. On some systems,
@@ -365,18 +494,32 @@ class Connection(object):
         :param int mode:
             int representation of octal mode for directory, default 777
 
-        :returns: nothing
+        :returns: None
 
         """
         self._sftp_connect()
         self._sftp.mkdir(remotepath, mode=int(str(mode), 8))
 
+    def normalize(self, remotepath):
+        """Return the expanded path, w.r.t the server, of a given path.  This
+        can be used to resolve symlinks or determine what the server believes
+        to be the :attr:`.pwd`, by passing '.' as remotepath.
+
+        :param str remotepath: path to be normalized
+
+        :return: (str) normalized form of the given path
+
+        :raises: IOError, if remotepath can't be resolved
+        """
+        self._sftp_connect()
+        return self._sftp.normalize(remotepath)
+
     def isdir(self, remotepath):
-        """return true if remotepath is a directory
+        """return true, if remotepath is a directory
 
         :param str remotepath: the path to test
 
-        :returns bool:
+        :returns: (bool)
 
         """
         self._sftp_connect()
@@ -391,7 +534,7 @@ class Connection(object):
 
         :param str remotepath: the path to test
 
-        :returns bool:
+        :returns: (bool)
 
         """
         self._sftp_connect()
@@ -408,7 +551,7 @@ class Connection(object):
         If remotedir already exists, silently complete. If a regular file is
         in the way, raise an exception.
 
-        :param str remotedir: the direcotry structure to create
+        :param str remotedir: the directory structure to create
         :param int mode:
             int representation of octal mode for directory, default 777
 
@@ -433,13 +576,25 @@ class Connection(object):
             if tail:
                 self.mkdir(remotedir, mode=mode)
 
+    def readlink(self, remotelink):
+        """Return the target of a symlink (shortcut).  The result will be
+        an absolute pathname.
+
+        :param str remotelink: remote path of the symlink
+
+        :return: (str) absolute path to target
+
+        """
+        self._sftp_connect()
+        return self._sftp.normalize(self._sftp.readlink(remotelink))
+
     def remove(self, remotefile):
         """remove the file @ remotefile, remotefile may include a path, if no
-        path, then cwd is used.  This method only works on files
+        path, then :attr:`.pwd` is used.  This method only works on files
 
         :param str remotefile: the remote file to delete
 
-        :returns: nothing
+        :returns: None
 
         :raises: IOError
 
@@ -447,12 +602,14 @@ class Connection(object):
         self._sftp_connect()
         self._sftp.remove(remotefile)
 
+    unlink = remove     # synonym for remove
+
     def rmdir(self, remotepath):
         """remove remote directory
 
         :param str remotepath: the remote directory to remove
 
-        :returns: nothing
+        :returns: None
 
         """
         self._sftp_connect()
@@ -465,7 +622,7 @@ class Connection(object):
 
         :param str remote_dest: the remote file/directory to put it
 
-        :returns: nothing
+        :returns: None
 
         :raises: IOError
 
@@ -478,7 +635,7 @@ class Connection(object):
 
         :param str remotepath: path to stat
 
-        :returns: SFTPAttributes object
+        :returns: (obj) SFTPAttributes
 
         """
         self._sftp_connect()
@@ -490,7 +647,7 @@ class Connection(object):
 
         :param str remotepath: path to stat
 
-        :returns: SFTPAttributes object
+        :returns: (obj) SFTPAttributes object
 
         """
         self._sftp_connect()
@@ -517,9 +674,9 @@ class Connection(object):
             mode (Python-style) to open file (always assumed binary)
         :param int bufsize: desired buffering (-1 = default buffer size)
 
-        :returns: an SFTPFile object representing the open file
+        :returns: (obj) SFTPFile, a handle the remote open file
 
-        :raises: IOError - if the file could not be opened.
+        :raises: IOError, if the file could not be opened.
 
         """
         self._sftp_connect()
@@ -530,7 +687,7 @@ class Connection(object):
 
         :param str remotepath: the remote path to verify
 
-        :returns bool: True if remotepath exists, else False
+        :returns: (bool) True, if remotepath exists, else False
 
         """
         self._sftp_connect()
@@ -546,7 +703,7 @@ class Connection(object):
 
         :param str remotepath: the remote path to verify
 
-        :returns bool: True if lexists, else False
+        :returns: (bool), True, if lexists, else False
 
         """
         self._sftp_connect()
@@ -572,6 +729,61 @@ class Connection(object):
         self._sftp_connect()
         self._sftp.symlink(remote_src, remote_dest)
 
+    def truncate(self, remotepath, size):
+        """Change the size of the file specified by path. Used to modify the
+        size of the file, just like the truncate method on Python file objects.
+        The new file size is confirmed and returned.
+
+        :param str remotepath: remote file path to modify
+        :param int|long size: the new file size
+
+        :returns: (int) new size of file
+
+        :raises: IOError, if file does not exist
+
+        """
+        self._sftp_connect()
+        self._sftp.truncate(remotepath, size)
+        return self._sftp.stat(remotepath).st_size
+
+    def walktree(self, remotepath, fcallback, dcallback, ucallback):
+        '''recursively descend, depth first, the directory tree rooted at
+        remotepath, calling discreet callback functions for each regular file,
+        directory and unknown file type.
+
+        :param str remotepath:
+            root of remote directory to descend, use '.' to start at
+            :attr:`.pwd`
+        :param callable fcallback:
+            callback function to invoke for a regular file.
+            (form: ``func(str)``)
+        :param callable dcallback:
+            callback function to invoke for a directory. (form: ``func(str)``)
+        :param callable ucallback:
+            callback function to invoke for an unknown file type.
+            (form: ``func(str)``)
+
+        :returns: None
+
+        :raises:
+
+        '''
+        self._sftp_connect()
+        for entry in self._sftp.listdir(remotepath):
+            pathname = os.path.join(remotepath, entry)
+            mode = self._sftp.stat(pathname).st_mode
+            if S_ISDIR(mode):
+                # It's a directory, call the dcallback function
+                dcallback(pathname)
+                # now, recurse into it
+                self.walktree(pathname, fcallback, dcallback, ucallback)
+            elif S_ISREG(mode):
+                # It's a file, call the fcallback function
+                fcallback(pathname)
+            else:
+                # Unknown file type
+                ucallback(pathname)
+
     @property
     def sftp_client(self):
         """give access to the underlying, connected paramiko SFTPClient object
@@ -580,7 +792,7 @@ class Connection(object):
 
         :params: None
 
-        :returns: the active SFTPClient object
+        :returns: (obj) the active SFTPClient object
 
         """
         self._sftp_connect()
@@ -591,7 +803,7 @@ class Connection(object):
         """Get tuple of currently used local and remote ciphers.
 
         :returns:
-            a tuple of currently used ciphers (local_cipher, remote_cipher)
+            (tuple of  str) currently used ciphers (local_cipher, remote_cipher)
 
         """
         return self._transport.local_cipher, self._transport.remote_cipher
@@ -600,11 +812,10 @@ class Connection(object):
     def security_options(self):
         """return the available security options recognized by paramiko.
 
-        :returns SecurityOptions:
-            a simple object security preferences of an
-            ssh transport. These are tuples of acceptable ciphers, digests,
-            key types, and key exchange algorithms, listed in order of
-            preference.
+        :returns:
+            (obj) security preferences of the ssh transport. These are tuples
+            of acceptable `.ciphers`, `.digests`, `.key_types`, and key exchange
+            algorithms `.kex`, listed in order of preference.
 
         """
 
@@ -614,11 +825,32 @@ class Connection(object):
     def logfile(self):
         '''return the name of the file used for logging or False it not logging
 
-        :returns: logfile(str) or False(bool)
+        :returns: (str)logfile or (bool) False
 
         '''
         return self._logfile
 
+    @property
+    def timeout(self):
+        ''' (float|None) get or set the underlying socket timeout for pending
+        read/write ops.  (default: None - no timeout)
+
+        :returns:
+            (float|None) seconds to wait for a pending read/write operation
+            before raising socket.timeout, or None for no timeout
+        '''
+        self._sftp_connect()
+        channel = self._sftp.get_channel()
+
+        return channel.gettimeout()
+
+    @timeout.setter
+    def timeout(self, val):
+        '''setter for timeout'''
+        self._sftp_connect()
+        channel = self._sftp.get_channel()
+        channel.settimeout(val)
+
     def __del__(self):
         """Attempt to clean up if not explicitly closed."""
         self.close()
@@ -636,7 +868,7 @@ def path_advance(thepath, sep=os.sep):
     :param str thepath: the path to navigate forwards
     :param str sep: the path separator to use, defaults to ``os.sep``
 
-    :returns generator: of strings
+    :returns: (iter)able of strings
 
     '''
     # handle a direct path
@@ -662,7 +894,7 @@ def path_retreat(thepath, sep=os.sep):
     :param str thepath: the path to retreat over
     :param str sep: the path separator to use, default to ``os.sep``
 
-    :returns generator: of strings
+    :returns: (iter)able of strings
 
     '''
     pre = ''
@@ -673,3 +905,20 @@ def path_retreat(thepath, sep=os.sep):
         if os.path.join(*parts):
             yield '%s%s' % (pre, os.path.join(*parts))
         parts = parts[:-1]
+
+def reparent(newparent, oldpath):
+    '''when copying or moving a directory structure, you need to re-parent the
+    oldpath.  When using os.path.join to calculate this new path, the
+    appearance of a / root path at the beginning of oldpath, supplants the
+    newparent and we don't want this to happen, so we need to make the oldpath
+    root appear as a child of the newparent.
+
+    :param: str newparent: the new parent location for oldpath (target)
+    :param str oldpath: the path being adopted by newparent (source)
+
+    :returns: (str) resulting adoptive path
+    '''
+
+    if oldpath[0] == os.sep:
+        oldpath = '.' + oldpath
+    return os.path.join(newparent, oldpath)
